import matplotlib.pyplot as plt


def plot_xic(process, ref_mz):
    fig, ax = plt.subplots(figsize=(15, 5), dpi=100)
    xic = process.loc[process["Mass"] == ref_mz].drop("Mass", axis=1)
    ax.plot(xic.index, xic["Intensity"], linewidth=0.5)
    plt.show()


def cut_data(filename: str, process, cut_range: tuple[int, int] | None = None):
    if cut_range is None:
        cut_range = input(f"Cut {filename} range (start, end): ").split(",")
    return process.loc[cut_range[0] : cut_range[1]]



import pandas as pd
import numpy as np
from scipy.stats import zscore


def max_normalize(mat: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Normalize data by max method.
    Args:
        mat: A dataframe.

    Returns:
        A normalized dataframe.

    """
    cell_max = mat.max(axis=1)
    return mat.div(cell_max, axis=0)


def mz_normalize(mat: pd.DataFrame, mz: float = 760.58, **kwargs) -> pd.DataFrame:
    """
    Normalize data by mz method.
    Args:
        mat: A dataframe.
        mz: The mz value to normalize.

    Returns:
        A normalized dataframe.

    """
    mz_df = mat.loc[:, mz]
    return mat.div(mz_df, axis=0)


def zscore_normalize(mat: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Normalize data by zscore method.
    Args:
        mat: A dataframe.

    Returns:
        A normalized dataframe.

    """
    return mat.apply(zscore)


def sum_normalize(mat: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """
    Normalize data by sum method.
    Args:
        mat: A dataframe.

    Returns:
        A normalized dataframe.

    """
    cell_sum = mat.sum(axis=1)
    return mat.div(cell_sum, axis=0)


def quantile_normalize(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """Perform Quantile-normalization on the given DataFrame.

    Args:
        df (pd.DataFrame): Input DataFrame with cells as rows and m/z values as columns

    Returns:
        pd.DataFrame: Quantile-normalized DataFrame
    """
    normalized_df = df.copy()
    normalized_df.fillna(0, inplace=True)

    for col in normalized_df.columns:
        sorted_col = normalized_df[col].sort_values(ascending=True)
        mean_rank = sorted_col.mean()
        sorted_index = sorted_col.index
        sorted_col_normalized = sorted_col.rank(method="average")

        col_normalized = pd.Series(index=sorted_index, data=sorted_col_normalized * mean_rank)
        col_normalized.sort_index(inplace=True)

        normalized_df[col] = col_normalized

    return normalized_df


def filter_mat(mat: pd.DataFrame, **kwargs) -> pd.DataFrame:
    # Drop all NaN rows
    mat = mat.dropna(how="all")
    return mat


def log_normalize(df: pd.DataFrame, base: float = 2, **kwargs) -> pd.DataFrame:
    """Perform log normalization on the given DataFrame.

    Args:
        df (pd.DataFrame): Input DataFrame with cells as rows and m/z values as columns
        base (float): Base of the logarithm (default is 2)

    Returns:
        pd.DataFrame: Log-normalized DataFrame
    """
    normalized_df = df.copy()
    normalized_df = normalized_df.applymap(lambda x: np.log(x + 1, dtype='float64') / np.log(base, dtype='float64') if x > 0 else 0)
    return normalized_df


FUNCTION_KEY = {
    "zscore": zscore_normalize,
    "max": max_normalize,
    "mz": mz_normalize,
    "sum": sum_normalize,
    "quantile": quantile_normalize,
    "log": log_normalize,
}


def normalize(data: pd.DataFrame, method: list[str], **kwargs) -> pd.DataFrame:
    """
    Normalize data by different methods.
    Args:
        data: A dataframe.
        method: The method to normalize data.
        **kwargs: The parameters of the method.

    Returns:
        A normalized dataframe.

    """
    for m in method:
        try:
            data = FUNCTION_KEY[m](data, **kwargs)
        except KeyError:
            raise KeyError(f"Method {m} is not supported.")
    return data





from configparser import ConfigParser
from pathlib import Path

import platform

from .default import DEFAULT_CONFIG

CONFIG_DIR = Path("~/.scmeta").expanduser()
CONFIG_PATH = CONFIG_DIR / "config.ini"
SYSTEM = platform.system()

def get_config():
    parser = ConfigParser()
    if not CONFIG_PATH.exists():
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        with open(CONFIG_PATH, "w") as f:
            f.write(DEFAULT_CONFIG)
    parser.read(CONFIG_PATH)
    return parser


class Parameters:
    def __init__(self, parameters):
        self.parameters = parameters

    def __getattr__(self, item):
        value = self.parameters[item]
        if value.isdigit():
            value = int(value)
        elif value.replace(".", "", 1).isdigit():
            value = float(value)
        elif value.lower() == "true":
            value = True
        elif value.lower() == "false":
            value = False
        elif value.lower() == "none":
            value = None
        return value

    def __getitem__(self, item):
        return self.__getattr__(item)


config = get_config()


PARAMETERS = Parameters(config["PARAMETERS"])
METABOLITE = Parameters(config["METABOLITE"])



from .knn import knn_impute


def fill_mat(mat, method: str):
    if method == "knn":
        return knn_impute(mat)
    elif method == "zero":
        return mat.fillna(0)
    elif method == "none":
        return mat
    else:
        raise KeyError("Invalid method")


from .load import PARAMETERS, SYSTEM
from .mz_filter import INCLUDE_LIST, EXCLUDE_LIST
import logging

# Create a logger
logging.basicConfig(
    format="[%(asctime)s][\033[32m%(levelname)s\033[0m] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
)



import requests

HMDB_API_URL = "https://hmdb.ca/metabolites/"
TOKEN = "5ZhBVFkIpe43Jh00ULwNIipGjfJCxrqpCnsvS1jj31sAFHoGg57vd46Tk95PhLfDVQ0J2vpb4kme86GldFkE5w%3D%3D"
header = {
    "Content-Type": "application/json",
    "Accept": "application/json",
    "Accept-Encoding": "gzip, deflate, br",
}


def gen_data(mz_list: list[float]):
    query = "\n".join([str(mz) for mz in mz_list])
    data = {
        "utf8": "✓",
        "authenticity_token": TOKEN,
        "query_masses": query,
        "ms_search_ion_mode": "positive",
        "adduct_type": "M+H",
        "tolerance_unit": "ppm",
        "ccs_predictor": "",
        "ccs_tolerance": 0.01,
        "commit": "Search",
    }
    return data


def get_hmdb_id(name):
    season = requests.post(HMDB_API_URL, data=gen_data([name]), headers=header)



from dataclasses import dataclass, field

import pandas as pd


@dataclass
class SCData:
    name: str
    raw: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())
    offset: float = 0
    process: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())
    mat: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())
    cell_pos: list = list[list[int]]
    cell_mat: pd.DataFrame = field(default_factory=lambda: pd.DataFrame())
    cell_count = property(lambda self: self.cell_mat.shape[0])
    database: bool = False

    def clear(self):
        self.mat = pd.DataFrame()

    def set_offset(self, offset: float | None):
        if self.offset is None or self.offset == 0:
            self.offset = offset
            self.raw["Mass"] += offset
        elif offset is None or offset == 0:
            pass
        else:
            _offset = offset - self.offset
            self.offset = offset
            self.raw["Mass"] += _offset

    def cut(self, start: float | None, end: float | None):
        self.raw = self.raw.loc[start:end]

    def xic(self, mz: float):
        return self.process.loc[self.process["Mass"] == mz]

    def show(self, _type="xic", **kwargs):
        pass

    def get_scan(self, scan: int, data_type: str = "raw"):
        if data_type == "raw":
            return self.raw.loc[scan]
        elif data_type == "process":
            return self.process.loc[scan]
        elif data_type == "mat":
            return self.mat.loc[scan]
        elif data_type == "cell_mat":
            for index, sublist in enumerate(self.cell_pos):
                if scan in sublist:
                    return self.cell_mat.loc[index]
        else:
            raise ValueError("data_type must be 'raw', 'process', 'mat' or 'cell_mat'")

    def reset(self):
        self.process = pd.DataFrame()
        self.mat = pd.DataFrame()
        self.cell_pos = []
        self.cell_mat = pd.DataFrame()
        self.raw["Mass"] -= self.offset
        self.offset = 0.0



import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.widgets import Button, RangeSlider


def cut_range_plot(lif: pd.DataFrame) -> (int, int):
    def f(x, range_):
        return x[range_[0] : range_[1]]

    init_start: int = 0
    init_end: int = len(lif["Time"])

    plt.style.use("fast")

    fig, axs = plt.subplots(2, 1, figsize=(10, 4))
    (line1,) = axs[0].plot(
        lif["Time"][init_start:init_end], lif["CH0"][init_start:init_end]
    )
    (line2,) = axs[1].plot(
        lif["Time"][init_start:init_end], lif["CH1"][init_start:init_end]
    )
    axs[0].set_title("CH0")
    axs[1].set_title("CH1")
    axs[1].set_xlabel("Time")
    fig.subplots_adjust(bottom=0.2)

    axrange = plt.axes([0.2, 0.075, 0.65, 0.03])
    srange = RangeSlider(
        axrange, "Range", 0, len(lif["Time"]), valinit=(init_start, init_end), valstep=1
    )

    def update(val):
        start = srange.val[0]
        end = srange.val[1]
        line1.set_data(lif["Time"][start:end], lif["CH0"][start:end])
        line2.set_data(lif["Time"][start:end], lif["CH1"][start:end])
        axs[0].set_ylim(
            [min(lif["CH0"][start:end]) - 500, max(lif["CH0"][start:end]) + 500]
        )
        axs[1].set_ylim(
            [min(lif["CH1"][start:end]) - 500, max(lif["CH1"][start:end]) + 500]
        )
        axs[0].set_xlim(
            [min(lif["Time"][start:end]) - 1, max(lif["Time"][start:end]) + 1]
        )
        axs[1].set_xlim(
            [min(lif["Time"][start:end]) - 1, max(lif["Time"][start:end]) + 1]
        )
        fig.canvas.draw_idle()

    srange.on_changed(update)

    resetax = fig.add_axes([0.8, 0.025, 0.1, 0.04])
    button_reset = Button(resetax, "Reset", color="mistyrose", hovercolor="0.975")

    def reset(event):
        srange.reset()

    button_reset.on_clicked(reset)

    saveax = fig.add_axes([0.7, 0.025, 0.1, 0.04])
    button_save = Button(saveax, "Save", color="lavender", hovercolor="0.975")

    def save(event):
        plt.close()

    button_save.on_clicked(save)
    plt.show()
    return int(srange.val[0]), int(srange.val[1])



import multiprocessing as mp
from functools import reduce

import time


class MultiProcessing:
    def __init__(self, cpu_num=mp.cpu_count()):
        self.cpu_num = cpu_num

    @staticmethod
    def process_sub_dict(args):
        sub_dict, func, args, kwargs = args
        result = []
        for key in sub_dict:
            processed = {key: func(sub_dict[key], *args, **kwargs)}
            result.append(processed)
        return result

    @staticmethod
    def merge_dicts(dict_list):
        result = {}
        for d in dict_list:
            result.update(d)
        return result

    def run(self, data, func, *args, **kwargs):
        num_processes = self.cpu_num

        # 将字典拆分成子字典列表
        data_items = list(data.items())
        sub_dicts = [dict(data_items[i::num_processes]) for i in range(num_processes)]

        with mp.Pool(num_processes) as pool:
            # 使用多进程处理子字典，得到一个包含 {key: value} 的列表
            processed_lists = pool.map(self.process_sub_dict, [(sub_dict, func, args, kwargs) for sub_dict in sub_dicts])

        # 合并处理过的列表
        result_list = []
        for plist in processed_lists:
            result_list.extend(plist)

        # 将包含 {key: value} 的列表合并成一个字典
        result_dict = self.merge_dicts(result_list)
        return result_dict


def key_func(x, num):
    return x * num

if __name__ == "__main__":
    data = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6, 'g': 7, 'h': 8}
    start = time.time()
    mpa = MultiProcessing()
    result = mpa.run(data, key_func, 2)
    print(result)



import os

from SCMeTA.accelerate import MultiThreader

from .thermo import load_thermo_data
from .process import load_process_data
from .waters import load_waters_data
from .format import *
from .database import load_from_database

FUNC_DICT = {
    "thermo": load_thermo_data,
    "process": load_process_data,
    "waters": load_waters_data,
}

SUFFIX_DICT = {
    "thermo": [".raw", ".txt"],
    "process": [".csv"],
    "waters": [".wiff", ".txt"],
}


def is_type(path: str, data_type: str) -> bool:
    suffix = os.path.splitext(path)[1]
    return suffix.lower() in SUFFIX_DICT[data_type]


def get_name_from_path(path: str) -> str:
    return os.path.splitext(os.path.basename(path))[0]


def path_from_database(path: str) -> str:
    base_path = ".Database/cyesi"
    return os.path.join(base_path, path)


def read_files_in_parallel(paths, names = None, data_type: str = "thermo") -> dict[str, SCData]:
    if names is None:
        names = [get_name_from_path(path) for path in paths]
    args = [(name, path) for name, path in zip(names, paths)]
    mt = MultiThreader()
    results = mt.run(FUNC_DICT[data_type], args)
    return results


def load_data(
    path: str | dict, name: str | None = None, data_type: str = "thermo"
) -> dict[str, SCData]:
    if isinstance(path, str):
        if os.path.isdir(path):
            files = os.listdir(path)
            paths = [os.path.join(path, file) for file in files if is_type(file, data_type)]
            return read_files_in_parallel(paths=paths, data_type=data_type)
        elif os.path.isfile(path) and is_type(path, data_type):
            if name is None:
                name = get_name_from_path(path)
            return {name: FUNC_DICT[data_type](name, path)}
        else:
            raise ValueError("Please provide a valid path")
    elif isinstance(path, dict):
        paths = [path_from_database(path) for path in path.values()]
        results = read_files_in_parallel(paths=paths, names=path.keys(), data_type=data_type)
        return results





import requests


class DatabaseConnector:
    def __init__(self, database_url: str = "https://database.xrzhanggroup.com"):
        self._session = requests.session()
        if "://" not in database_url:
            self.url = f"http://{database_url}"
        else:
            self.url = database_url

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._session.close()

    def get_path(self, file_id: int):
        response = self._session.get(f"{self.url}/api/v1/data/path/{file_id}").json()
        return response

    def get_paths(self, file_ids: list[int]):
        return [self.get_path(file_id) for file_id in file_ids]


def load_from_database(file_ids: list[int] | int):
    with DatabaseConnector() as db:
        if isinstance(file_ids, int):
            data = db.get_path(file_ids)
        elif isinstance(file_ids, list):
            data = db.get_paths(file_ids)
        else:
            raise ValueError("Please provide a valid file id or a list of file ids")
    return data




import threading


class MultiThreader:
    def __init__(self, n_threads=10):
        self.n_threads = n_threads

    def run(self, func, args_list):
        results = {}
        threads = []

        def worker(*args):
            result = func(*args)
            results[args[0]] = result

        for args in args_list:
            thread = threading.Thread(target=worker, args=args)
            threads.append(thread)

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        return results



import numpy as np

import pandas as pd
from matplotlib.patches import Circle, RegularPolygon
from matplotlib.path import Path
from matplotlib.projections.polar import PolarAxes
from matplotlib.projections import register_projection
from matplotlib.spines import Spine
from matplotlib.transforms import Affine2D

from matplotlib.figure import Figure, Axes


def radar_factory(num_vars, frame="circle"):
    """Create a radar chart with `num_vars` axes.

    This function creates a RadarAxes projection and registers it.

    Parameters
    ----------
    num_vars : int
        Number of variables for radar chart.
    frame : {'circle' | 'polygon'}
        Shape of frame surrounding axes.

    """
    # calculate evenly-spaced axis angles
    theta = np.linspace(0, 2 * np.pi, num_vars, endpoint=False)

    class RadarAxes(PolarAxes):
        name = "radar"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # rotate plot such that the first axis is at the top
            self.set_theta_zero_location("N")

        def fill(self, *args, closed=True, **kwargs):
            """Override fill so that line is closed by default"""
            return super().fill(closed=closed, *args, **kwargs)

        def plot(self, *args, **kwargs):
            """Override plot so that line is closed by default"""
            lines = super().plot(*args, **kwargs)
            for line in lines:
                self._close_line(line)

        def _close_line(self, line):
            x, y = line.get_data()
            # FIXME: markers at x[0], y[0] get doubled-up
            if x[0] != x[-1]:
                x = np.concatenate((x, [x[0]]))
                y = np.concatenate((y, [y[0]]))
                line.set_data(x, y)

        def set_varlabels(self, labels):
            self.set_thetagrids(np.degrees(theta), labels)

        def _gen_axes_patch(self):
            # The Axes patch must be centered at (0.5, 0.5) and of radius 0.5
            # in axes coordinates.
            if frame == "circle":
                return Circle((0.5, 0.5), 0.5)
            elif frame == "polygon":
                return RegularPolygon((0.5, 0.5), num_vars, radius=0.5, edgecolor="k")
            else:
                raise ValueError("unknown value for 'frame': %s" % frame)

        def draw(self, renderer):
            """Draw. If frame is polygon, make gridlines polygon-shaped"""
            if frame == "polygon":
                gridlines = self.yaxis.get_gridlines()
                for gl in gridlines:
                    gl.get_path()._interpolation_steps = num_vars
            super().draw(renderer)

        def _gen_axes_spines(self):
            if frame == "circle":
                return super()._gen_axes_spines()
            elif frame == "polygon":
                # spine_type must be 'left'/'right'/'top'/'bottom'/'circle'.
                spine = Spine(
                    axes=self,
                    spine_type="circle",
                    path=Path.unit_regular_polygon(num_vars),
                )
                # unit_regular_polygon gives a polygon of radius 1 centered at
                # (0, 0) but we want a polygon of radius 0.5 centered at (0.5,
                # 0.5) in axes coordinates.
                spine.set_transform(
                    Affine2D().scale(0.5).translate(0.5, 0.5) + self.transAxes
                )

                return {"polar": spine}
            else:
                raise ValueError("unknown value for 'frame': %s" % frame)

    register_projection(RadarAxes)
    return theta


def radar(
    mat_list: list[pd.DataFrame],
    mz_list: dict[float, str],
    ax: Axes,
    fig: Figure,
    title: str = "Radar",
):
    mz_keys = mz_list.keys()
    cell_features = []
    for mat in mat_list:
        cell_features.append(mat.mean(axis=0).loc[mz_keys].values)
    cell_features = np.array(cell_features)
    N = len(mz_list)
    theta = radar_factory(N, frame="polygon")
    spoke_labels = list[mz_list.values()]
    ax.set_rgrids([0.2, 0.4, 0.6, 0.8])
    ax.set_title(
        title,
        weight="bold",
        size="medium",
        position=(0.5, 1.1),
        horizontalalignment="center",
        verticalalignment="center",
    )
    for d, color in zip(cell_features, ["b", "r", "g", "m", "y"]):
        ax.plot(theta, d, color=color)
        ax.fill(theta, d, facecolor=color, alpha=0.25)
    ax.set_varlabels(spoke_labels)
    return fig



import pandas as pd

from .format import SCData


def load_process_data(name, path) -> SCData:
    process = pd.read_csv(path, index_col=0)
    data = SCData(name)
    data.process = process
    return data



import pandas as pd
from combat.pycombat import pycombat

from SCMeTA.file import SCData


def combat_batch_correction(data: dict[str, SCData], tags: list[str]) -> dict[str, SCData]:
    # 获取细胞组名和数据框列表
    d_list = []
    for key, value in data.items():
        cell_mat = value.cell_mat.T
        for tag in tags:
            if key in tag:
                cell_mat.columns = [tag for _ in range(len(cell_mat.columns))]
        d_list.append(value.cell_mat.T)

    # 合并所有细胞组数据
    df_expression = pd.concat(d_list, join="inner", axis=1)

    # 创建批次标签列表
    batch = []
    datasets = d_list
    for j in range(len(datasets)):
        batch.extend([j for _ in range(len(datasets[j].columns))])

    # run pyComBat
    df_corrected = pycombat(df_expression, batch)

    start = 0
    end = 0
    # 获取校正后的各细胞组数据
    for key in data.keys():
        d = df_corrected.T
        end += len(data[key].cell_mat.index)
        d = d.iloc[start:end, :]
        d.index = data[key].cell_mat.index
        start = end
        data[key].cell_mat = d
    return data



import numpy as np
import pandas as pd
from matplotlib.figure import Axes


def line_df(data: pd.DataFrame, ax: Axes):
    x, y1, y2 = data.index, data.iloc[:, 0], data.iloc[:, 1]
    x_label, y_label = data.index.name, "Intensity"
    ax[0].plot(x, y1, color="black", linewidth=0.8)
    ax[1].plot(x, y2, color="red", linewidth=0.8)
    ax[0].set_title(data.columns[0])
    ax[1].set_title(data.columns[1])
    return x_label, y_label


def line(data: pd.DataFrame | np.ndarray, ax: Axes | np.ndarray):
    if isinstance(data, pd.DataFrame):
        x_label, y_label = line_df(data, ax)
    elif isinstance(data, np.ndarray):
        if data.ndim == 1:
            x, y = np.arange(len(data)), data
        elif data.ndim == 2:
            x, y = data[:, 0], data[:, 1]
        elif data.T.ndim == 2:
            x, y = data.T[:, 0], data.T[:, 1]
        else:
            raise ValueError("data must be 1D or 2D")
        x_label, y_label = "x", "y"
        ax.plot(x, y, color="black", linewidth=0.8)
    else:
        raise TypeError(f"data type {type(data)} is not supported")
    return x_label, y_label


if __name__ == "__main__":
    path = "/test.csv"
    pd.read_csv(path, index_col=0)



import pandas as pd
import numpy as np


def tfidf(mat: pd.DataFrame) -> pd.DataFrame:
    data = mat.replace(0, np.nan)
    total_intensity = data.sum(axis=1)
    tf = data.divide(total_intensity, axis=0)
    idf = np.log(data.shape[0] / data.count())
    tfidf = tf * idf
    return tfidf


if __name__ == "__main__":
    mat = pd.read_csv(
        "/Users/Estrella/Developer/CyESI/Process/Data/idf-test.csv", index_col=0
    )
    result = tfidf(mat)
    # loc 760.58



import numpy as np
from scipy.spatial.distance import cdist
from scipy.stats import ttest_ind, f_oneway


def euclidean_distance(data_list1: np.ndarray, data_list2: np.ndarray):
    euclidean_dist = cdist(data_list1, data_list2, 'euclidean')
    return euclidean_dist


# 计算两组数据间的曼哈顿距离
def manhattan_distance(data_list1: np.ndarray, data_list2: np.ndarray):
    manhattan_dist = cdist(data_list1, data_list2, 'cityblock')
    return manhattan_dist


def t_test(data1: np.ndarray, data2: np.ndarray):
    t, p = ttest_ind(data1, data2)
    return t, p


def anova_test(data_list: list[np.ndarray]):
    f, p = f_oneway(*data_list)
    return f, p
# t-检验比较两组数据是否有显著差异


if __name__ == '__main__':
    pass


from matplotlib.figure import Axes
import pandas as pd

COLOR_KEY = [
    "firebrick",
    "darkorange",
    "gold",
    "greenyellow",
    "deepskyblue",
    "navy",
    "violet",
    "magenta",
    "pink",
    "black",
]


def box(data: dict[str, pd.DataFrame], h_statistic: float, p_value: float, ax: Axes):
    # 绘制箱线图
    values = [df.iloc[:, 0].T.to_numpy() for df in data.values()]
    labels = [name for name in data.keys()]
    ax.boxplot(values, labels=labels, showmeans=True, meanline=True, patch_artist=True)

    # 设定颜色
    for patch, color in zip(ax.artists, COLOR_KEY):
        patch.set_facecolor(color)

    # 设置图表标题和坐标轴标签
    ax.set_title(f"Kruskal-Wallis H Test Result")
    ax.set_xlabel('Data Group')
    ax.set_ylabel('Value')



import pandas as pd


def find_cell_index(xic: pd.DataFrame, max_ratio=0.1) -> list:
    max_intensity = xic.max()
    cell_region = xic[xic > max_ratio * max_intensity].index
    return cell_region.to_list()


def find_adjacent(cell_array: list) -> list[list]:
    cell_group = []
    for i in cell_array:
        if not cell_group or i != cell_group[-1][-1] + 1:
            cell_group.append([i])
        else:
            cell_group[-1].append(i)
    return cell_group


def find_cell(
    mat: pd.DataFrame, refer_mz: float = 760.58, max_ratio: float = 0.1
) -> list[list]:
    """Find the cell regions.

    Parameters
    ----------
    mat : pd.DataFrame
        The raw data.
    refer_mz : float, optional
        The reference m/z, by default 760.58
    max_ratio : float, optional

    Returns
    -------
    list[list]
        The cell regions.
    """
    xic = mat.loc[:, refer_mz].fillna(0)
    cell_array = find_cell_index(xic, max_ratio=max_ratio)
    return find_adjacent(cell_array)


def merge_cell(
    mat: pd.DataFrame, cell_pos: list[list], adjacent: int = 3
) -> pd.DataFrame:
    """Combine the cell regions.

    Parameters
    ----------
    mat : pd.DataFrame
        The raw data.
    cell_pos : list[list]
        The cell regions.
    adjacent : int, optional
        The adjacent cell regions, by default 3

    Returns
    -------
    pd.DataFrame
        The combined cell regions.
    """
    cell_series_list = []
    index_list = []
    for index, group in enumerate(cell_pos):
        if len(group) <= adjacent:
            cell_series_list.append(mat.loc[group].sum())
            index_list.append(index)
    cell_mat = pd.concat(cell_series_list, axis=1).T
    # Set index
    cell_mat.index = index_list
    cell_mat.index.name = "Scan"
    return cell_mat



import sqlite3

from SCMeTA.config import PARAMETERS


class Connector:
    def __init__(self):
        self.conn = sqlite3.connect(PARAMETERS.dbpath)
        self.cursor = self.conn.cursor()

    def __enter__(self):
        self.conn = sqlite3.connect(PARAMETERS.dbpath)
        self.cursor = self.conn.cursor()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def search(self, mz: float, rl: float = 1.008, tol: float = 0.01):
        mass = mz - rl
        self.cursor.execute(
            f"SELECT * FROM hmdb WHERE monisotopic_molecular_weight BETWEEN {mass - tol} AND {mass + tol}"
        )
        return self.cursor.fetchall()

    def search_list(self, mz_list: list[float], rl: float = 1.008, tol: float = 0.01) -> dict[float, list]:
        return {mz: self.search(mz, rl, tol) for mz in mz_list}





import os
import sys
import numpy as np
import pandas as pd
import tqdm

from pythonnet import load
load("coreclr")

import clr
from System import String

dir_path = os.path.dirname(os.path.realpath(__file__))

sys.path.append(os.path.join(dir_path, "dll"))


clr.AddReference("ThermoFisher.CommonCore.Data")
clr.AddReference("ThermoFisher.CommonCore.RawFileReader")
import ThermoFisher
from ThermoFisher.CommonCore.Data.Interfaces import IScanEventBase, IScanEvent

'''
rawFile = ThermoFisher.CommonCore.RawFileReader.RawFileReaderAdapter.FileFactory(raw_filename)
var scanStatistics = rawFile.GetScanStatsForScanNumber(1);
var seg = rawFile.GetSegmentedScanFromScanNumber(1, scanStatistics);
var scanEvent = rawFile.GetScanEventForScanNumber(1);
var trailerData = rawFile.GetTrailerExtraInformation(1);
'''


def DotNetArrayToNPArray(arr, dtype):
    return np.array(list(arr), dtype=dtype)


def NpArrayToDataFrame(arr, columns: list[str]):
    return pd.DataFrame(arr.transpose(), columns=columns)


'''
APIs are similar to pymsfilereader(https://github.com/frallain/pymsfilereader), but some APIs have not be implemented yet."
'''


class RawFileReader(object):
    # static class members
    sampleType = {0: 'Unknown',
                  1: 'Blank',
                  2: 'QC',
                  3: 'Standard Clear (None)',
                  4: 'Standard Update (None)',
                  5: 'Standard Bracket (Open)',
                  6: 'Standard Bracket Start (multiple brackets)',
                  7: 'Standard Bracket End (multiple brackets)'}

    controllerType = {-1: 'No device',
                      0: 'MS',
                      1: 'Analog',
                      2: 'A/D card',
                      3: 'PDA',
                      4: 'UV',
                      'No device': -1,
                      'MS': 0,
                      'Analog': 1,
                      'A/D card': 2,
                      'PDA': 3,
                      'UV': 4}

    massAnalyzerType = {'ITMS': 0,
                        'TQMS': 1,
                        'SQMS': 2,
                        'TOFMS': 3,
                        'FTMS': 4,
                        'Sector': 5,
                        0: 'ITMS',
                        1: 'TQMS',
                        2: 'SQMS',
                        3: 'TOFMS',
                        4: 'FTMS',
                        5: 'Sector'}
    activationType = {'CID': 0,
                      'MPD': 1,
                      'ECD': 2,
                      'PQD': 3,
                      'ETD': 4,
                      'HCD': 5,
                      'Any activation type': 6,
                      'SA': 7,
                      'PTR': 8,
                      'NETD': 9,
                      'NPTR': 10,
                      'UVPD': 11,
                      'ETHCD': 12,  # not Thermo's build-in activation types
                      'ETCID': 13,  # not Thermo's build-in activation types
                      0: 'CID',
                      1: 'MPD',
                      2: 'ECD',
                      3: 'PQD',
                      4: 'ETD',
                      5: 'HCD',
                      6: 'Any activation type',
                      7: 'SA',
                      8: 'PTR',
                      9: 'NETD',
                      10: 'NPTR',
                      11: 'UVPD',
                      12: 'ETHCD',  # not Thermo's build-in activation types
                      13: 'ETCID',  # not Thermo's build-in activation types
                      }

    detectorType = {'Valid': 0,
                    'Any': 1,
                    'NotValid': 2,
                    0: 'Valid',
                    1: 'Any',
                    2: 'NotValid',
                    }

    scanDataType = {'Centroid': 0,
                    'Profile': 1,
                    'Any': 2,
                    0: 'Centroid',
                    1: 'Profile',
                    2: 'Any',
                    }

    scanType = {'Full': 0,
                'Zoom': 1,
                'SIM': 2,
                'SRM': 3,
                'CRM': 4,
                'Any': 5,
                'Q1MS': 6,
                'Q3MS': 7,
                0: 'Full',
                1: 'SIM',
                2: 'Zoom',
                3: 'SRM',
                4: 'CRM',
                5: 'Any',
                6: 'Q1MS',
                7: 'Q3MS',
                }

    def __init__(self, filename, **kwargs):
        self.filename = os.path.abspath(filename)
        self.filename = os.path.normpath(self.filename)

        self.source = ThermoFisher.CommonCore.RawFileReader.RawFileReaderAdapter.FileFactory(self.filename)

        if not self.source.IsOpen:
            raise IOError(
                "RAWfile '{0}' could not be opened, is the file accessible ?".format(
                    self.filename))
        self.source.SelectInstrument(ThermoFisher.CommonCore.Data.Business.Device.MS, 1)

        self.StartTime = self.GetStartTime()
        self.EndTime = self.GetEndTime()
        self.FirstSpectrumNumber = self.GetFirstSpectrumNumber()
        self.LastSpectrumNumber = self.GetLastSpectrumNumber()
        self.LowMass = self.GetLowMass()
        self.HighMass = self.GetHighMass()
        self.MassResolution = self.GetMassResolution()
        self.NumSpectra = self.GetNumSpectra()

    def Close(self):
        """Closes a raw file and frees the associated memory."""
        self.source.Dispose()

    def GetFileName(self):
        """Returns the fully qualified path name of an open raw file."""
        return self.source.FileName

    def GetCreatorID(self):
        """Returns the creator ID. The creator ID is the
        logon name of the user when the raw file was acquired."""
        return self.source.CreatorId

    def GetCreationDate(self):
        """Returns the file creation date in DATE format."""
        # https://msdn.microsoft.com/en-us/library/82ab7w69.aspx
        # The DATE type is implemented using an 8-byte floating-point number
        return self.source.CreationDate.ToOADate()

    def IsError(self):
        """Returns the error state flag of the raw file. A return value of TRUE indicates that an error has
        occurred. For information about the error, call the GetErrorCode or GetErrorMessage
        functions."""
        return self.source.IsError

    def IsThereMSData(self):
        """This function checks to see if there is MS data in the raw file. A return value of TRUE means
        that the raw file contains MS data. You must open the raw file before performing this check."""
        return self.source.HasMsData

    def InAcquisition(self):
        """Returns the acquisition state flag of the raw file. A return value of TRUE indicates that the
        raw file is being acquired or that all open handles to the file during acquisition have not been
        closed."""
        return self.source.InAcquisition

    def RefreshViewOfFile(self):
        """Refreshes the view of a file currently being acquired. This function provides a more efficient
        mechanism for gaining access to new data in a raw file during acquisition without closing and
        reopening the raw file. This function has no effect with files that are not being acquired."""
        return self.source.RefreshViewOfFile

    def GetExpectedRunTime(self):
        """Gets the expected acquisition run time for the current controller. The actual acquisition may
        be longer or shorter than this value. This value is intended to allow displays to show the
        expected run time on chromatograms. To obtain an accurate run time value during or after
        acquisition, use the GetEndTime function."""
        return self.source.ExpectedRunTime

    def GetNumTrailerExtra(self):
        """Gets the trailer extra entries recorded for the current controller. Trailer extra entries are only
        supported for MS device controllers and are used to store instrument specific information for
        each scan if used."""
        return self.source.RunHeaderEx.TrailerExtraCount

    def GetMaxIntegratedIntensity(self):
        """Gets the highest integrated intensity of all the scans for the current controller. This value is
        only relevant to MS device controllers."""
        return self.source.RunHeaderEx.MaxIntegratedIntensity

    def GetMaxIntensity(self):
        """Gets the highest base peak of all the scans for the current controller. This value is only relevant
        to MS device controllers."""
        return self.source.RunHeaderEx.MaxIntensity

    def GetComment1(self):
        """Returns the first comment for the current controller. This value is typically only set for raw
        files converted from other file formats."""
        return self.source.RunHeaderEx.Comment1

    def GetComment2(self):
        """Returns the second comment for the current controller. This value is typically only set for raw
        files converted from other file formats."""
        return self.source.RunHeaderEx.Comment2

    def GetFilters(self):
        """Returns the list of unique scan filters for the raw file. This function is only supported for MS
        device controllers."""
        return list(self.source.GetFilters())

    # INSTRUMENT BEGIN
    def GetInstName(self):
        """Returns the instrument name, if available, for the current controller."""
        return String.Join(" -> ", self.source.GetAllInstrumentNamesFromInstrumentMethod())

    # INSTRUMENT END

    def GetScanEventStringForScanNum(self, scanNumber):
        """This function returns scan event information as a string for the specified scan number."""
        return self.source.GetScanEventStringForScanNumber(scanNumber)

    def GetNumberOfMassRangesFromScanNum(self, scanNumber):
        """This function gets the number of MassRange data items in the scan."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).MassRangeCount

    def GetMassRangeFromScanNum(self, scanNumber, massRangeIndex):
        """This function retrieves information about the mass range data of a scan (high and low
        masses). You can find the count of mass ranges for the scan by calling
        GetNumberOfMassRangesFromScanNum()."""
        range = IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetMassRange(massRangeIndex)
        return range.Low, range.High

    def GetNumberOfSourceFragmentsFromScanNum(self, scanNumber):
        """This function gets the number of source fragments (or compensation voltages) in the scan."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).SourceFragmentationInfoCount

    def GetSourceFragmentValueFromScanNum(self, scanNumber, sourceFragmentIndex):
        """This function retrieves information about one of the source fragment values of a scan. It is
        also the same value as the compensation voltage. You can find the count of source fragments
        for the scan by calling GetNumberOfSourceFragmentsFromScanNum ()."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetSourceFragmentationInfo(
            sourceFragmentIndex)

    def GetIsolationWidthForScanNum(self, scanNumber, MSOrder=0):
        """This function returns the isolation width for the scan specified by scanNumber and the
        transition specified by MSOrder (0 for MS1?) from the scan event structure in the raw file."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetIsolationWidth(MSOrder)

    def GetCollisionEnergyForScanNum(self, scanNumber, MSOrder=0):
        """This function returns the collision energy for the scan specified by scanNumber and the
        transition specified by MSOrder (0 for MS1?) from the scan event structure in the raw file. """
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetEnergy(MSOrder)

    def GetActivationTypeForScanNum(self, scanNumber, MSOrder=0):
        """This function returns the activation type for the scan specified by scanNumber and the
        transition specified by MSOrder from the scan event structure in the RAW file.
        The value returned in the pnActivationType variable is one of the following:
        CID  0
        MPD 1
        ECD  2
        PQD 3
        ETD 4
        HCD 5
        Any activation type 6
        SA 7
        PTR 8
        NETD 9
        NPTR 10
        UVPD 11"""
        return RawFileReader.activationType[
            IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetActivation(MSOrder)]

    def GetMassAnalyzerTypeForScanNum(self, scanNumber):
        """This function returns the mass analyzer type for the scan specified by scanNumber from the
        scan event structure in the RAW file. The value of scanNumber must be within the range of
        scans or readings for the current controller. The range of scans or readings for the current
        controller may be obtained by calling GetFirstSpectrumNumber and
        GetLastSpectrumNumber.
        return RawFileReader.massAnalyzerType[IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).MassAnalyzer]"""

    def GetDetectorTypeForScanNum(self, scanNumber):
        """This function returns the detector type for the scan specified by scanNumber from the scan
        event structure in the RAW file."""
        return RawFileReader.detectorType[IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).Detector]

    def GetNumberOfMassCalibratorsFromScanNum(self, scanNumber):
        """This function gets the number of mass calibrators (each of which is a double) in the scan."""
        return IScanEvent(self.source.GetScanEventForScanNumber(scanNumber)).MassCalibratorCount

    def GetMassCalibrationValueFromScanNum(self, scanNumber, massCalibrationIndex):
        """This function retrieves information about one of the mass calibration data values of a scan.
        You can find the count of mass calibrations for the scan by calling
        GetNumberOfMassCalibratorsFromScanNum()."""
        return IScanEvent(self.source.GetScanEventForScanNumber(scanNumber)).GetMassCalibrator(massCalibrationIndex)

    def GetMassResolution(self):
        """Gets the mass resolution value recorded for the current controller. The value is returned as one
        half of the mass resolution. For example, a unit resolution controller returns a value of 0.5.
        This value is only relevant to scanning controllers such as MS."""
        return self.source.RunHeaderEx.MassResolution

    def GetLowMass(self):
        """Gets the lowest mass or wavelength recorded for the current controller. This value is only
        relevant to scanning devices such as MS or PDA."""
        return self.source.RunHeaderEx.LowMass

    def GetHighMass(self):
        """Gets the highest mass or wavelength recorded for the current controller. This value is only
        relevant to scanning devices such as MS or PDA."""
        return self.source.RunHeaderEx.HighMass

    def GetStartTime(self):
        """Gets the start time of the first scan or reading for the current controller. This value is typically
        close to zero unless the device method contains a start delay."""
        return self.source.RunHeaderEx.StartTime

    def GetEndTime(self):
        return self.source.RunHeaderEx.EndTime

    def GetNumSpectra(self):
        """Gets the number of spectra acquired by the current controller. For non-scanning devices like
        UV detectors, the number of readings per channel is returned."""
        return self.source.RunHeaderEx.SpectraCount

    def GetFirstSpectrumNumber(self):
        """Gets the first scan or reading number for the current controller. If data has been acquired, this
        value is always one."""
        return self.source.RunHeaderEx.FirstSpectrum

    def GetLastSpectrumNumber(self):
        """Gets the last scan or reading number for the current controller."""
        return self.source.RunHeaderEx.LastSpectrum

    def ScanNumFromRT(self, RT):
        """Returns the closest matching scan number that corresponds to RT for the current controller.
        For non-scanning devices, such as UV, the closest reading number is returned. The value of
        RT must be within the acquisition run time for the current controller. The acquisition run
        time for the current controller may be obtained by calling GetStartTime and GetEndTime."""
        return self.source.ScanNumberFromRetentionTime(RT)

    def ScanNumFromRTInSeconds(self, RTInSeconds):
        """Returns the closest matching scan number that corresponds to RT for the current controller.
        For non-scanning devices, such as UV, the closest reading number is returned. The value of
        RT must be within the acquisition run time for the current controller. The acquisition run
        time for the current controller may be obtained by calling GetStartTime and GetEndTime."""
        return self.ScanNumFromRT(RTInSeconds / 60)

    def RTFromScanNum(self, scanNumber):
        """Returns the closest matching run time or retention time that corresponds to scanNumber for
        the current controller. For non-scanning devices, such as UV, the scanNumber is the reading
        number."""
        return self.source.RetentionTimeFromScanNumber(scanNumber)

    def RTInSecondsFromScanNum(self, scanNumber):
        """Returns the closest matching run time or retention time that corresponds to scanNumber for
        the current controller. For non-scanning devices, such as UV, the scanNumber is the reading
        number."""
        return self.RTFromScanNum(scanNumber) * 60

    def IsProfileScanForScanNum(self, scanNumber):
        """Returns TRUE if the scan specified by scanNumber is a profile scan, FALSE if the scan is a
        centroid scan."""
        return not self.source.GetScanStatsForScanNumber(scanNumber).IsCentroidScan

    def IsCentroidScanForScanNum(self, scanNumber):
        """Returns TRUE if the scan specified by scanNumber is a centroid scan, FALSE if the scan is a
        profile scan."""
        return self.source.GetScanStatsForScanNumber(scanNumber).IsCentroidScan

    def GetMSOrderForScanNum(self, scanNumber):
        """This function returns the MS order for the scan specified by scanNumber from the scan
        event structure in the raw file.
        The value returned in the pnScanType variable is one of the following:
        Neutral gain -3
        Neutral loss -2
        Parent scan -1
        Any scan order 0
        MS  1
        MS2  2
        MS3  3
        MS4  4
        MS5  5
        MS6  6
        MS7  7
        MS8  8
        MS9  9
        """
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).MSOrder

    def GetNumberOfMSOrdersFromScanNum(self, scanNumber):
        """This function gets the number of MS reaction data items in the scan event for the scan
        specified by scanNumber and the transition specified by MSOrder from the scan event
        structure in the raw file."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).MassCount

    def GetPrecursorMassForScanNum(self, scanNumber, MSOrder=0):
        """This function returns the precursor mass for the scan specified by scanNumber and the
        transition specified by MSOrder (0 for precursor in MS1?) from the scan event structure in the RAW file."""
        return IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber)).GetReaction(MSOrder).PrecursorMass

    def GetPrecursorRangeForScanNum(self, scanNumber, MSOrder=0):
        """This function returns the first and last precursor mass values of the range and whether they are
        valid for the scan specified by scanNumber and the transition specified by MSOrder (0 for precursor in MS1?) from the scan event structure in the raw file."""
        scanEvent = IScanEventBase(self.source.GetScanEventForScanNumber(scanNumber))
        firstMass = scanEvent.GetFirstPrecursorMass(MSOrder)
        lastMass = scanEvent.GetLastPrecursorMass(MSOrder)
        return firstMass, lastMass

    def GetBasePeakForScanNum(self, scanNumber):
        """This function returns the base peak mass and intensity of mass spectrum."""
        stat = self.source.GetScanStatsForScanNumber(scanNumber)
        return stat.BasePeakMass, stat.BasePeakIntensity

    # "View/Scan header", lower part
    def GetTrailerExtraForScanNum(self, scanNumber):
        """Returns the recorded trailer extra entry labels and values for the current controller. This
        function is only valid for MS controllers.
        NOTE : XCALIBUR INTERFACE "View/Scan header", lower part
        """
        trailerData = self.source.GetTrailerExtraInformation(scanNumber)
        return dict(zip(trailerData.Labels, trailerData.Values))

    def GetProfileMassListFromScanNum(self, scanNumber):
        scanStatistics = self.source.GetScanStatsForScanNumber(scanNumber)
        segmentedScan = self.source.GetSegmentedScanFromScanNumber(scanNumber, scanStatistics)
        return np.array([DotNetArrayToNPArray(segmentedScan.Positions, float),
                         DotNetArrayToNPArray(segmentedScan.Intensities, float)])

    def GetCentroidMassListFromScanNum(self, scanNumber):
        scan = ThermoFisher.CommonCore.Data.Business.Scan.FromFile(self.source, scanNumber)
        scanStatistics = self.source.GetScanStatsForScanNumber(scanNumber)
        if scanStatistics.IsCentroidScan:
            segmentedScan = self.source.GetSegmentedScanFromScanNumber(scanNumber, scanStatistics)
            scan = pd.DataFrame(np.array([DotNetArrayToNPArray(segmentedScan.Positions, float),
                                          DotNetArrayToNPArray(segmentedScan.Intensities, float)]).transpose(),
                                columns=['Mass', 'Intensity'])
        elif scan.HasCentroidStream:
            stream = self.source.GetCentroidStream(scanNumber, False)
            scan = pd.DataFrame(np.array([DotNetArrayToNPArray(stream.Masses, float),
                                          DotNetArrayToNPArray(stream.Intensities, float)]).transpose(),
                                columns=['Mass', 'Intensity'])
        else:
            print("Profile scan {0} cannot be centroided!".format(scanNumber))
            segmentedScan = self.source.GetSegmentedScanFromScanNumber(scanNumber, scanStatistics)
            scan = pd.DataFrame(np.array([DotNetArrayToNPArray(segmentedScan.Positions, float),
                                          DotNetArrayToNPArray(segmentedScan.Intensities, float)]).transpose(),
                                columns=['Mass', 'Intensity'])
        scan["Scan"] = scanNumber
        return scan

    def GetFullScanMassList(self):
        """This function returns the mass list of full scan."""
        scan_range = range(self.FirstSpectrumNumber, self.LastSpectrumNumber + 1)
        scan_mass_list = [self.GetCentroidMassListFromScanNum(scan) for scan in tqdm.tqdm(scan_range)]
        return pd.concat(scan_mass_list).set_index("Scan")





# from communities.algorithms import spectral_clustering
#
#
# class Communities:
#     def __init__(self):
#         pass
#
#     def clean(self):
#         pass
#
#     def clustering(self):
#         pass



from bokeh.io import output_notebook

from SCMeTA.plot.Bokeh import heatmap
from SCMeTA.file import SCData


class BokehPlot:
    def __init__(self, data: dict[str, SCData]):
        self.__data = [d.cell_mat for d in data.values()]
        self.__data_count = len(data)
        self.__cell_range = {
            key: value.cell_mat.shape[0] for key, value in data.items()
        }
        output_notebook()

    def heatmap(self, title: str = "Heatmap", **kwargs):
        heatmap(mat_list=self.__data, title=title)



import os
import logging

import pandas as pd

from SCMeTA.file import SCData
from SCMeTA.method import (
    filter_occ,
    to_mat,
    to_list,
    find_cell,
    merge_cell,
    noise_subtract,
    filter_assem,
    filter_mat,
    normalize,
    round_columns,
)
from SCMeTA.batch import combat_batch_correction
from SCMeTA.method.fill import fill_mat
from SCMeTA.file import load_data, load_from_database
from SCMeTA.config import PARAMETERS
from SCMeTA.accelerate import MultiProcessing

logger = logging.getLogger(__name__)


def _filter_occ(data: SCData, resolution: float, count: int):
    data.process = filter_occ(data.raw.copy(), resolution, count)


class Process:
    def __init__(
        self, ref_mz: float = 760.58, mz1: float = 760.58, mz2: float = 732.55
    ):
        self.data: dict[str, SCData] = {}

        self.__mz1: float = mz1
        self.__mz2: float = mz2
        self.ref_mz: float = ref_mz
        self.__dir = None
        self.__mp = MultiProcessing()
        pass

    def load(
        self,
        path: str,
        file_name: str | None = None,
        data_type: str = "thermo",
    ):
        """
        Add a file to the MSProcess
        Args:
            path: File path or directory path.
            file_name: File Name, cannot work if path is a directory.
            data_type: Data type, thermo_raw file / water wiff file / processed csv file are supported.
        """
        self.data.update(load_data(path, file_name, data_type))
        self.__dir = os.path.dirname(path)

    def load_database(self, file_id: int | list[int]):
        """
        Load data from database.
        Args:
            file_id: File ID in the database.

        Returns:

        """
        data = load_from_database(file_id)
        for key, value in data.items():
            self.data.update(load_data(value, key, "database"))

    def load_processed(self, path, file_name: str | None = None, file_type: str = "cell_mat"):
        """
        Load processed data.
        Args:
            path: File path or directory path.
            file_name: File Name, cannot work if path is a directory.
            file_type:

        Returns:

        """
        if os.path.isdir(path):
            for file in os.listdir(path):
                if file.endswith(".csv"):
                    file_name = os.path.basename(file).split(".")[0]
                    data = pd.read_csv(os.path.join(path, file), index_col=0)
                    data.columns = [float(i) for i in data.columns]
                    self.data[file_name] = SCData(name=file_name, cell_mat=data)
        else:
            data = pd.read_csv(path, index_col=0)
            data.columns = [float(i) for i in data.columns]
            self.data[file_name] = SCData(name=file_name, cell_mat=data)

    def filter_occ(
        self,
        count: int = PARAMETERS.count,
        resolution: float = PARAMETERS.resolution,
        file_name: str | None = None,
    ):
        """
        Filter out the data with low occurrence
        Args:
            count: Minimum occurrence of the data, default 10
            resolution: Resolution of the data, default 0.01
            file_name: File name, if None, all files will be processed.
        """
        if file_name is None:
            # self.__mp.run(self.data, _filter_occ, resolution, count)
            for ms_data in self.data.values():
                ms_data.process = filter_occ(ms_data.raw.copy(), resolution, count)
        else:
            self.data[file_name].process = filter_occ(
                self.data[file_name].raw.copy(), resolution, count
            )
        logger.info("Filter out the data with low occurrence.")

    def gen_mat(self, file_name: str | None = None):
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.mat = to_mat(ms_data.process)
        else:
            self.data[file_name].mat = to_mat(self.data[file_name].process)

    def denoise(self, max_ratio: float = PARAMETERS.maxratio, file_name: str | None = None):
        """
        Find cell and subtract noise.
        Args:
            max_ratio: If the ratio of the max value to the second max value is larger than this value, the cell will be
            file_name: File name, if None, all files will be processed.
        """
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.cell_pos = find_cell(
                    ms_data.mat, self.ref_mz, max_ratio=max_ratio
                )
                ms_data.mat = noise_subtract(ms_data.mat, ms_data.cell_pos)
        else:
            ms_data = self.data[file_name]
            ms_data.cell_pos = find_cell(
                ms_data.mat, self.ref_mz, max_ratio=max_ratio
            )
            ms_data.mat = noise_subtract(ms_data.mat, ms_data.cell_pos)
            self.data[file_name] = ms_data
        logger.info("Noise subtracted!")

    def merge_cell(self, adjacent: int = PARAMETERS.adjacent, file_name: str | None = None):
        """
        Combine the adjacent cells
        Args:
            adjacent: Number of adjacent cells to be combined, default 3.
            If number is larger than the number of cells, the data will be dropped.
            file_name: File name, if None, all files will be processed.
        """
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.cell_mat = merge_cell(
                    ms_data.mat, ms_data.cell_pos, adjacent=adjacent
                )
        else:
            self.data[file_name].cell_mat = merge_cell(
                self.data[file_name].mat,
                self.data[file_name].cell_pos,
                adjacent=adjacent,
            )

    def filter_assem(self, snr: float = PARAMETERS.snr, file_name: str | None = None):
        """
        Filter out the data with Intensity/Noise < snr
        Args:
            snr: Minimum SNR of the data, default 3.0
            file_name: File name, if None, all files will be processed.
        """
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.cell_mat = filter_assem(
                    ms_data.mat, ms_data.cell_mat, ms_data.cell_pos, snr
                )
        else:
            self.data[file_name].cell_mat = filter_assem(
                self.data[file_name].mat,
                self.data[file_name].cell_mat,
                self.data[file_name].cell_pos,
                snr,
            )
        logger.info("Filter out the data with low SNR.")

    def round_mat(
        self, resolution: float = PARAMETERS.resolution, file_name: str | None = None
    ):
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.mat = round_columns(ms_data.mat, resolution)
        else:
            self.data[file_name].mat = round_columns(
                self.data[file_name].mat, resolution
            )

    def gen_process(self, file_name: str | None = None):
        if file_name is None:
            for ms_data in self.data.values():
                ms_data.process = to_list(ms_data.cell_mat, ms_data.cell_pos)
        else:
            self.data[file_name].process = to_list(
                self.data[file_name].cell_mat, self.data[file_name].cell_pos
            )

    def filter_mat(
        self,
        threshold: float = PARAMETERS.threshold,
        name_list: list[str] | None = None,
        lock_mz: bool = PARAMETERS.lock,
        method: str = "all"
    ):
        """
        Filter out the data with low occurrence
        Args:
            threshold: Minimum occurrence rate of the data, default 0.2
            name_list: File name list, if None, all files will be processed.
            lock_mz: If True, the mz you select in lock mz file will be locked, default False.
            method: Method to filter the data, default "all", can be "all", "any", "none".
        """
        if name_list is None:
            name_list = list(self.data.keys())
        mat_list = [self.data[name].cell_mat for name in name_list]
        total_mat = filter_mat(mat_list, threshold, lock_mz, method)
        for index, name in enumerate(name_list):
            self.data[name].cell_mat = total_mat[index]
        logger.info("Filter out the mat data.")

    def normalize(self,
                  data: dict[str, SCData],
                  normalize_method: list[str],
                  file_name: str | None = None):
        if file_name is None:
            for ms_data in data.values():
                ms_data.cell_mat = normalize(
                    ms_data.cell_mat, normalize_method, mz=self.ref_mz
                )
        else:
            data[file_name].cell_mat = normalize(
                data[file_name].cell_mat, normalize_method, mz=self.ref_mz
            )
        logger.info("Normalization finished.")

    def fill(self,
             data: dict[str, SCData],
             file_name: str | None = None,
             fillna_method: str = "knn"
             ):
        if file_name is None:
            for ms_data in data.values():
                ms_data.cell_mat = fill_mat(ms_data.cell_mat, fillna_method)
        else:
            data[file_name].cell_mat = fill_mat(
                data[file_name].cell_mat, fillna_method
            )
        logger.info("Fillna finished.")

    def combat(self, data: dict[str, SCData], tag_list: list[str], file_name: str | None = None):
        data = combat_batch_correction(data)
        return data

    def info(self, file_name: str | None = None):
        """
        Print the information of the MSProcess
        Args:
            file_name: File name, if None, all files' info will be printed.
        """
        if file_name is None:
            for ms_data in self.data.values():
                logger.info(
                    " ".join(
                        [
                            f"File name: {ms_data.name}",
                            f"Cells count: {ms_data.cell_count}",
                            f"Peaks count: {ms_data.cell_mat.shape[1]}",
                        ]
                    )
                )
        else:
            ms_data = self.data[file_name]
            logger.info(
                " ".join(
                    [
                        f"File name: {ms_data.name}",
                        f"Cells count: {ms_data.cell_count}",
                        f"Peaks count: {ms_data.cell_mat.shape[1]}",
                    ]
                )
            )

    def clear_memory(self, file_name: str | None = None):
        """
        Clear the memory of the MSProcess
        Args:
            file_name: File name, if None, all files will be processed.
        """
        if file_name is None:
            for value in self.data.values():
                value.clear()
        else:
            self.data[file_name].clear()
        logger.info("Memory cleared")

    def save(self, file_name: str | None = None, data_type: str = "mat", path: str | None = None):
        """
        Save the MSProcess
        Args:
            file_name: File name, if None, all files will be processed.
            data_type: File type, default "mat", "process" is also available.
            path: Path to save the file, default None,
             which means the file will be saved in the same directory as the data you loaded.
        """
        if path is None:
            dir_path = os.path.join(self.__dir, "Process")
        else:
            dir_path = path

        def get_data_by_type(ms_data, data_type):
            if data_type == "cell_mat":
                return ms_data.cell_mat
            if data_type == "mat":
                return ms_data.mat
            elif data_type == "process":
                return ms_data.process
            else:
                raise ValueError("Type should be 'mat' or 'process'")

        if not os.path.exists(dir_path):
            os.mkdir(dir_path)
        if file_name is None:
            for ms_data in self.data.values():
                get_data_by_type(ms_data, data_type).to_csv(
                    os.path.join(dir_path, f"{ms_data.name}_{data_type}.csv")
                )
        else:
            get_data_by_type(self.data[file_name], data_type).to_csv(
                os.path.join(dir_path, f"{file_name}_{data_type}.csv")
            )
        logger.info("Data saved!")

    def pre_process(
        self,
        file_name: str | list[str] | None = None,
        offset: float | None = None,
        cut_range: tuple[int, int] | None = None,
        resolution: float = PARAMETERS.resolution,
        count: int = PARAMETERS.count,
    ):
        """
        Pre-process the data, including offset, cut, and round.
        Args:
            file_name: File name
            offset: Offset of the data, if None, it will ignore the offset step.
            cut_range: Cut range of the data, if None, it will ignore the cut step.
            resolution: Resolution of the data, default 0.01.
            count: Minimum occurrence of the data, default 10.
        Returns:

        """
        if isinstance(file_name, str):
            self.data[file_name].set_offset(offset=offset)
            self.data[file_name].cut(start=cut_range[0], end=cut_range[1])
        elif isinstance(file_name, list):
            for name in file_name:
                offset = float(input(f"Please input the offset of {name}: "))
                cut_range = (
                    int(input(f"Please input the start of cut range of {name}: ")),
                    int(input(f"Please input the end of cut range of {name}: ")),
                )
                self.data[name].set_offset(offset=offset)
                self.data[name].cut(start=cut_range[0], end=cut_range[1])
        elif file_name is None:
            pass
        self.filter_occ(resolution=resolution, count=count)
        logger.info("Pre-process finished.")

    def process(
            self,
            max_ratio: float = PARAMETERS.maxratio,
            adjacent: int = PARAMETERS.adjacent,
            snr: float = PARAMETERS.snr,
            resolution: float = PARAMETERS.resolution,
            threshold: float = PARAMETERS.threshold,
            lock_mz: bool = PARAMETERS.lock,
            filter_method: str = "all"
    ):
        """
        Args:
            max_ratio: If the ratio of the max value to the second max value is larger than this value, the cell will be
            adjacent: The number of adjacent cells to be combined
            snr: Signal to noise ratio
            resolution: Resolution of the data
            threshold: Threshold of the data
            lock_mz: If True, the mz you select in lock mz file will be locked
            filter_method: Method of filtering
        """
        if self.data is None:
            logger.warning("Please check the data carefully!")
            raise ValueError("No data loaded, please load data first")
        self.gen_mat()
        self.round_mat(resolution=resolution)
        self.denoise(max_ratio=max_ratio)
        self.merge_cell(adjacent=adjacent)
        self.filter_assem(snr=snr)
        self.filter_mat(threshold=threshold, lock_mz=lock_mz, method=filter_method)
        self.info()
        self.clear_memory()
        return self.data

    def post_process(
            self,
            data: dict[str, SCData] = None,
            normalize_method=None,
            fillna_method: str = "none",
            tags: list[str] = None,
    ):
        """
        Args:
            data: Dict of MSData
            normalize_method: Method of normalization
            fillna_method: Method of filling nan
            tags: Tags of the data
        Returns:
            Dict of MSData
        """
        if normalize_method is None:
            normalize_method = ["mz"]
        if data is None:
            data = self.data
        self.normalize(data=data, normalize_method=normalize_method)
        self.fill(data=data, fillna_method=fillna_method)
        # self.combat()
        return data



import pandas as pd

from .load import METABOLITE, PARAMETERS

if PARAMETERS.lock:
    if METABOLITE.online:
        INCLUDE_LIST = pd.read_csv(METABOLITE.include)["mz"].tolist()
        EXCLUDE_LIST = pd.read_csv(METABOLITE.exclude)["mz"].tolist()
    else:
        INCLUDE_LIST = pd.read_csv(METABOLITE.include)["mz"].tolist()
        EXCLUDE_LIST = pd.read_csv(METABOLITE.exclude)["mz"].tolist()
else:
    INCLUDE_LIST = None
    EXCLUDE_LIST = None



import pandas as pd

from bokeh.layouts import column
from bokeh.models import RangeTool
from bokeh.plotting import figure, show


def range_tool(process: pd.DataFrame, refer_mz: float) -> (int, int):
    xic = process.loc[process["Mass"] == refer_mz]
    xic.reset_index(inplace=True)
    start = xic["Scan"].min()
    end = xic["Scan"].max()
    p = figure(
        height=300,
        width=800,
        tools="xpan",
        toolbar_location=None,
        x_axis_type="datetime",
        x_axis_location="above",
        background_fill_color="#efefef",
        x_range=(start, end),
    )
    p.line("Scan", "Intensity", source=xic)
    p.yaxis.axis_label = "Intensity"

    select = figure(
        title="Drag the middle and edges of the selection box to change the range above",
        height=130,
        width=800,
        y_range=p.y_range,
        x_axis_type="datetime",
        y_axis_type=None,
        tools="",
        toolbar_location=None,
        background_fill_color="#efefef",
    )

    rt = RangeTool(x_range=p.x_range)
    rt.overlay.fill_color = "navy"
    rt.overlay.fill_alpha = 0.2

    select.line("Scan", "Intensity", source=xic)
    select.ygrid.grid_line_color = None
    select.add_tools(rt)
    select.toolbar.active_multi = rt

    show(column(p, select))



import numpy as np
import pandas as pd


def discriminate(
    data_list: dict[str, pd.DataFrame], method: str, n_components: int = 2
) -> np.ndarray:
    """
    Discriminate data by different methods.
    Args:
        data_list: A dict of dataframes or a MSProcess object.
        method: The method to discriminate data.
        n_components: The number of components to keep.

    Returns:
        A numpy array of the result of the method.
    """
    from sklearn.manifold import TSNE
    from sklearn.decomposition import PCA, KernelPCA
    from umap import UMAP

    METHODS = {
        "pca": PCA,
        "kpca": KernelPCA,
        "tsne": TSNE,
        "umap": UMAP,
    }
    full = pd.concat(data_list.values(), axis=0).fillna(0)
    reduce = METHODS[method](n_components=n_components)
    return reduce.fit_transform(full)


def to_mat(
        full: np.ndarray, cell_range: dict[str, int], n_components: int = 2
) -> dict[str, pd.DataFrame]:
    """
    Convert the result of discriminate to a dataframe.
    Args:
        full: The result of discriminate.
        cell_range: The range of each cell.
        n_components

    Returns:
        A dataframe of the result of discriminate.
    """
    data_list = {}
    start = 0
    end = 0
    col_label = ["x", "y", "z"]
    col = col_label[:n_components]
    for cell, cell_count in cell_range.items():
        end += cell_count
        value = full[start:end, 0:n_components]
        data_list[cell] = pd.DataFrame(
            value, columns=col, index=range(cell_count)
        )
        start += cell_count
    return data_list



from .filter import filter_occ, filter_mat
from .to_mat import to_mat, to_list
from .cell import find_cell, merge_cell
from .noise import noise_subtract, filter_assem
from .normalize import normalize
from .round import round_columns, round_rows
from .combine_mat import combine_mat
from .cut import cut_data
from .k_w_test import k_w_test

from .machine_learning import discriminate



import numpy as np
import pandas as pd

from SCMeTA.config import INCLUDE_LIST, EXCLUDE_LIST


def sum_df(df, scan) -> pd.DataFrame:
    df = df.groupby("Mass").sum().reset_index()
    df.insert(0, "Scan", scan)
    df = df.set_index("Scan")
    return df


def peaks_combine(_raw: pd.DataFrame, resolution: float = 0.01) -> pd.DataFrame:
    try:
        _raw["Mass"] = _raw["Mass"].divide(0.01).apply(np.floor).mul(0.01)
    except AttributeError:
        _raw["Mass"] = np.floor(_raw["Mass"] / resolution) * resolution
    max_scan = _raw.index.max()
    min_scan = _raw.index.min()
    temp = [sum_df(_raw.loc[scan], scan) for scan in range(min_scan, max_scan + 1)]
    return pd.concat(temp)


def filter_occ(
    raw: pd.DataFrame, resolution: float = 0.01, count: int = 10
) -> pd.DataFrame:
    """
    Filter out peaks that occur less than count times in the data.
    :param raw: Raw data.
    :param resolution: Resolution of m/z.
    :param count: Minimum number of occurrences.
    :return: List of filtered peaks.
    """
    process = peaks_combine(raw, resolution)
    peaks = process["Mass"].value_counts()
    peaks = peaks[peaks >= count]
    process = process[process["Mass"].isin(peaks.index)]
    return process


def filter_mat(
    mat_list: list[pd.DataFrame], threshold: float = 0.2, lock: bool = False, method: str = "all"
) -> list[pd.DataFrame]:
    """Filter matrix by count
    Args:
        mat_list (list[pd.DataFrame]): List of matrix
        threshold (float, optional): Threshold. Defaults to 0.2.
        lock (bool, optional): Whether to lock the mz list. Defaults to False.
    Returns:
        pd.DataFrame: Filtered matrix
    """
    if method == "all":
        count = sum([mat.shape[0] for mat in mat_list]) * threshold
        large_mat = pd.concat(mat_list)
        large_mat.sort_index(axis=1, inplace=True)
        mz_bool = large_mat.count() >= count
        if lock:
            include_bool = large_mat.columns.isin(INCLUDE_LIST)
            exclude_bool = large_mat.columns.isin(EXCLUDE_LIST)
            mz_bool = mz_bool | include_bool
            mz_bool = mz_bool & ~exclude_bool
        mz_list = mz_bool[mz_bool].index
        # if mat in mat_list doesn't have the mz, it will be filled with NaN
        mat_list = [mat.reindex(columns=mz_list, fill_value=None) for mat in mat_list]
        mat_list = [mat[mz_list] for mat in mat_list]
        return mat_list
    elif method == "any":
        index_lists = []
        for mat in mat_list:
            count = mat.shape[0] * threshold
            mz_bool = mat.count() >= count
            if lock:
                include_bool = mz_bool.index.isin(INCLUDE_LIST)
                exclude_bool = mz_bool.index.isin(EXCLUDE_LIST)
                mz_bool = mz_bool | include_bool
                mz_bool = mz_bool & ~exclude_bool
            index_lists.append(mz_bool[mz_bool])
        mz_bool = pd.concat(index_lists, axis=1)
        mz_list = mz_bool.index
        new_mat_list = []
        for mat in mat_list:
            mat = mat.reindex(columns=mz_list, fill_value=None)
            mat = mat[mz_list]
            new_mat_list.append(mat)
        return new_mat_list


if __name__ == '__main__':
    # func any threshold
    # func all threshold
    pass


import pandas as pd
import numpy as np
from scipy import stats
from matplotlib.figure import Axes


def clean_df(
    df0: pd.DataFrame, df1: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame, list]:
    df0.dropna(how="all", inplace=True, axis=1)
    df1.dropna(how="all", inplace=True, axis=1)
    col_0 = df0.columns.tolist()
    col_1 = df1.columns.tolist()
    columns = [c for c in col_0 if c in col_1]
    df0 = df0.loc[:, columns].fillna(0)
    df1 = df1.loc[:, columns].fillna(0)
    return df0, df1, columns


def volcano(
    mat1: pd.DataFrame,
    mat2: pd.DataFrame,
    name1: str,
    name2: str,
    ax: Axes,
    x_threshold: float = 1.0,
    y_threshold: float = 1.30103,
):
    df0, df1, columns = clean_df(mat1, mat2)

    _list1 = []
    _list0 = np.log2((df0.mean(axis=0) / df1.mean(axis=0)).tolist())
    for i in columns:  # 算差异的显著性
        _, p = stats.ttest_ind(df0.loc[:, i], df1.loc[:, i])
        # 做t-test，我省略了方差齐性检验
        _list1.append(-np.log(p))
        # 得到y轴数据的list

    df_volcano = pd.DataFrame({"Mass": columns, "x": _list0, "y": _list1})
    # 构造用于做火山图的dataframe

    # 分组染色
    df_volcano["group"] = "black"  # 增加一列，索引为group
    df_volcano.loc[
        (df_volcano.x > x_threshold) & (df_volcano.y > y_threshold), "group"
    ] = "tab:red"  # 右上角为红色
    df_volcano.loc[
        (df_volcano.x < -x_threshold) & (df_volcano.y > y_threshold), "group"
    ] = "tab:blue"  # 左上角为蓝色
    df_volcano.loc[df_volcano.y < y_threshold, "group"] = "dimgrey"  # 阈值以下点为灰色

    # 设置坐标轴范围
    x_min, x_max = -2, 2
    y_min, y_max = 0, 20

    # 绘制散点图
    ax.set(xlim=(x_min, x_max), ylim=(y_min, y_max), title="")  # 设置坐标轴的长度
    ax.scatter(df_volcano["x"], df_volcano["y"], s=2, c=df_volcano["group"])  # 设置标记大小
    ax.set_xlabel("log2(fold change)")
    ax.set_ylabel("-log10(P)")
    ax.spines["right"].set_visible(False)  # 去掉右边框
    ax.spines["top"].set_visible(False)  # 去掉上边框

    # 水平和竖直线
    ax.vlines(
        -x_threshold, y_min, y_max, color="dimgrey", linestyle="dashed", linewidth=1
    )  # 画竖直线
    ax.vlines(
        x_threshold, y_min, y_max, color="dimgrey", linestyle="dashed", linewidth=1
    )  # 画竖直线
    ax.hlines(
        y_threshold, x_min, x_max, color="dimgrey", linestyle="dashed", linewidth=1
    )  # 画竖水平线

    ax.set_xticks(range(-2, 3, 1))  # 设置x轴刻度起点和步长
    ax.set_yticks(range(0, 21, 5))  # 设置y轴刻度起点和步长

    # 把红点和蓝点标上Mass
    index0 = df_volcano[df_volcano.group == "tab:red"].index.tolist()
    index1 = df_volcano[df_volcano.group == "tab:blue"].index.tolist()
    for i in index0:
        ax.annotate(
            columns[i],
            xy=(_list0[i], _list1[i]),
            xytext=(_list0[i] + 0.02, _list1[i] + 0.02),
            size=2,
        )
        # 调整字号和位置
    for i in index1:
        ax.annotate(
            columns[i],
            xy=(_list0[i], _list1[i]),
            xytext=(_list0[i] + 0.02, _list1[i] + 0.02),
            size=2,
        )

    title = f"{name1} and {name2}"
    ax.set_title(title)



import os
import logging

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.figure import Figure, Axes

from SCMeTA.plot.Mpl import scatter, heatmap, line, radar, volcano, bar, box
from SCMeTA.method import round_rows, round_columns, k_w_test
from SCMeTA.method.machine_learning import discriminate, to_mat
from SCMeTA.file import SCData

logger = logging.getLogger(__name__)

FIGURE_SIZE = {
    "heatmap": (7, 5),
    "scatter": (5, 5),
    "line": (10, 5),
}


class MplPlot:
    def __init__(self):
        self.__data: dict[str, SCData] | None = None
        self.__mat: dict[str, pd.DataFrame] = {}
        self.__cell_range: dict[str, int] = {}

    def __read_csv(self, path: str):
        mat = pd.read_csv(path, index_col=0)
        name = os.path.basename(path).split(".")[0]
        self.__mat[name] = mat
        self.__cell_range[name] = mat.shape[0]
        logger.info(f"Successfully load {name}.")

    def load(self, data: dict[str, SCData] | None = None, path: str | None = None):
        if path is not None:
            # Load Cell Mat from dir or file
            if os.path.isdir(path):
                files = os.listdir(path)
                for file in files:
                    if file.endswith(".csv"):
                        self.__read_csv(os.path.join(path, file))
            elif os.path.isfile(path):
                self.__read_csv(path)
        elif data is not None:
            self.__data = data
            for name, ms_data in data.items():
                self.__mat[name] = ms_data.cell_mat
            logger.info("Successfully load data")
        else:
            raise ValueError("Please provide data or path")
        self.__cell_range = {key: mat.shape[0] for key, mat in self.__mat.items()}

    @staticmethod
    def init_plot(rows: int, cols: int, figure_type: str) -> tuple[Figure, Axes]:
        """
        Initialize a matplotlib figure and axes
        Args:
            rows: number of rows
            cols: number of columns
            figure_type: figure type
        Returns:
            fig: matplotlib figure
            ax: matplotlib axes
        """
        figsize = FIGURE_SIZE[figure_type]
        fig, ax = plt.subplots(
            nrows=rows,
            ncols=cols,
            figsize=(figsize[0] * cols, figsize[1] * rows),
            dpi=100,
        )
        return fig, ax

    def pca(self, ax=None, n_components: int = 2):
        full = discriminate(self.__mat, method="pca", n_components=n_components)
        if ax is None:
            fig, ax = self.init_plot(1, 1, "scatter")
        scatter(data=full, cell_range=self.__cell_range, ax=ax, title="PCA")
        return to_mat(full, self.__cell_range)

    def tsne(self, ax=None, n_components: int = 2):
        full = discriminate(self.__mat, method="tsne", n_components=n_components)
        if ax is None:
            fig, ax = self.init_plot(1, 1, "scatter")
        scatter(data=full, cell_range=self.__cell_range, ax=ax, title="t-SNE")
        return to_mat(full, self.__cell_range)

    def umap(self, ax=None, n_components: int = 2):
        full = discriminate(self.__mat, method="umap", n_components=n_components)
        if ax is None:
            fig, ax = self.init_plot(1, 1, "scatter")
        scatter(data=full, cell_range=self.__cell_range, ax=ax, title="UMAP")
        return to_mat(full, self.__cell_range)

    def scatter_select(self, method: list[str] | None = None, n_components: int = 2):
        if method is None:
            method = ["pca", "tsne", "umap"]
        fig, ax = self.init_plot(1, len(method), "scatter")
        for i, m in enumerate(method):
            if m == "pca":
                self.pca(ax[i], n_components)
            elif m == "tsne":
                self.tsne(ax[i], n_components)
            elif m == "umap":
                self.umap(ax[i], n_components)
            else:
                raise ValueError(f"method {m} is not supported")

    def heatmap(self, ax=None, fig=None, **kwargs):
        if ax is None:
            fig, ax = self.init_plot(1, 1, "heatmap")
        heatmap(
            self.__mat,
            cell_range=self.__cell_range,
            ax=ax,
            fig=fig,
            title="Heatmap",
            **kwargs,
        )

    def volcano(self, name1: str, name2: str, ax=None, **kwargs):
        mat1 = self.__mat[name1]
        mat2 = self.__mat[name2]
        if ax is None:
            fig, ax = self.init_plot(1, 1, "scatter")
        volcano(mat1, mat2, name1, name2, ax=ax, **kwargs)

    def box(self, ax=None, method: str = "tsne"):
        if ax is None:
            fig, ax = self.init_plot(1, 1, "scatter")
        full_data = discriminate(self.__mat, method=method, n_components=1)
        dict_data = to_mat(full_data, self.__cell_range, n_components=1)
        h_statistic, p_value = k_w_test(dict_data)
        box(dict_data, ax=ax, h_statistic=h_statistic, p_value=p_value)
        return dict_data

    # def radar(self, ax=None, mz_list: pd.DataFrame | None = None):
    #     if ax is None:
    #         fig, ax = self.init_plot(1, 1)
    #         radar(self.__mat, ax, self.__cell_range, title="Radar")
    #         return fig
    #     radar(self.__mat, ax, self.__cell_range, title="Radar")

    def ms_compare(self, name: str):
        fig, ax = self.init_plot(2, 1, "line")
        ms_data = self.__data[name]
        scan_list = ms_data.cell_pos.to_list()
        scan = int(input(f"Please Select a scan from:\n {scan_list}"))
        if scan not in scan_list:
            logger.warning(f"Scan {scan} is not in the scan list")
            raise ValueError(f"Scan {scan} is not in the scan list")
        line1 = round_columns(ms_data.get_scan(scan, data_type="raw"), axis=0)
        line2 = ms_data.get_scan(scan, data_type="cell_mat")
        full = pd.concat([line1, line2], axis=1).fillna(0)
        full.columns = ["Raw", "Process"]
        bar(full, ax=ax)
        return full

    def cell_compare(self, name: str):
        fig, ax = self.init_plot(2, 1, "line")
        ms_data = self.__data[name]
        cell_list = ms_data.cell_mat.index.tolist()
        cell1 = int(input(f"Please Select a cell from:\n {cell_list}"))
        cell2 = int(input(f"Please Select a cell from:\n {cell_list}"))
        if cell1 not in cell_list or cell2 not in cell_list:
            logger.warning(f"Cell {cell1} or {cell2} is not in the cell list")
            raise ValueError(f"Cell {cell1} or {cell2} is not in the cell list")
        line1 = ms_data.cell_mat.loc[cell1]
        line2 = ms_data.cell_mat.loc[cell2]
        full = pd.concat([line1, line2], axis=1).fillna(0)
        full.columns = [cell1, cell2]
        bar(full, ax=ax)
        return full



import pandas as pd
import numpy as np


def round_columns(mat: pd.DataFrame, resolution: float = 0.01, axis=1) -> pd.DataFrame:
    """Round the values of a column to n decimal places.

    Args:
        mat (pd.DataFrame): The matrix to round.
        resolution (float, optional): The resolution to round to. Defaults to 0.01.
        axis (int, optional): The axis to round. Defaults to 1.
    Returns:
        pd.DataFrame: The rounded matrix.
    """
    n = int(np.log10(1 / resolution))
    if axis == 1:
        c = mat.columns.to_numpy(dtype=float)
        c = np.round(c, n)
        mat.columns = c
    elif axis == 0:
        c = mat.index.to_numpy(dtype=float)
        c = np.round(c, n)
        mat.index = c
    return mat


def round_rows(
    mat: pd.DataFrame, key: str = "Mass", resolution: float = 0.01
) -> pd.DataFrame:
    """Round the values of a row to n decimal places.

    Args:
        mat (pd.DataFrame): The matrix to round.
        key (str): The key to round.
        resolution (float, optional): The resolution to round to. Defaults to 0.01.

    Returns:
        pd.DataFrame: The rounded matrix.
    """
    n = int(np.log10(1 / resolution))
    c = mat.loc[:, key].to_numpy(dtype=float)
    c = np.round(c, n)
    mat.loc[:, key] = c
    mat = mat.groupby(key).sum()
    return mat



from sklearn.impute import KNNImputer
import pandas as pd


def knn_impute(data):
    # Remove non numeric columns
    data = data.dropna(axis=1, how="all")
    imputer = KNNImputer(n_neighbors=5)
    imputed_data = imputer.fit_transform(data)
    imputed_data_df = pd.DataFrame(imputed_data)
    imputed_data_df.columns = data.columns
    imputed_data_df.index = data.index
    return imputed_data_df



import pandas as pd


def to_mat(data: pd.DataFrame) -> pd.DataFrame:
    mat = data.pivot_table(
        index="Scan",
        columns="Mass",
        values="Intensity",
    ).fillna(0)
    return mat


def to_list(mat: pd.DataFrame, cell_pos: list[list]) -> pd.DataFrame:
    index_list = mat.index
    index = [cell_pos[i][0] for i in index_list]
    mat.index = index
    mat = mat.stack().reset_index()
    mat.columns = ["Scan", "Mass", "Intensity"]
    mat.set_index("Scan", inplace=True)
    return mat



import numpy as np
import pandas as pd
from matplotlib.widgets import PolygonSelector
from matplotlib.path import Path
import matplotlib.pyplot as plt


class SelectFromCollection:
    def __init__(self, ax, collection, alpha_other=0.3):
        self.canvas = ax.figure.canvas
        self.collection = collection
        self.alpha_other = alpha_other

        self.xys = collection.get_offsets()
        self.Npts = len(self.xys)

        # Ensure that we have separate colors for each object
        self.fc = collection.get_facecolors()
        if len(self.fc) == 0:
            raise ValueError("Collection must have a facecolor")
        elif len(self.fc) == 1:
            self.fc = np.tile(self.fc, (self.Npts, 1))

        self.poly = PolygonSelector(ax, self.onselect)
        self.ind = []

    def onselect(self, verts):
        path = Path(verts)
        self.ind = np.nonzero(path.contains_points(self.xys))[0]
        self.fc[:, -1] = self.alpha_other
        self.fc[self.ind, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()

    def disconnect(self):
        self.poly.disconnect_events()
        self.fc[:, -1] = 1
        self.collection.set_facecolors(self.fc)
        self.canvas.draw_idle()


def polygonize(
    mat_list: list[pd.DataFrame], m1=732.55, m2=760.58
) -> list[pd.DataFrame]:
    mat = pd.concat(mat_list).reset_index()
    range_list = [mat.shape[0] for mat in mat_list]
    fig, ax = plt.subplots()
    pts = ax.scatter(mat.loc[:, m1], mat.loc[:, m2])
    selector = SelectFromCollection(ax, pts)
    fig.show()
    selector.disconnect()
    select_list = []
    for index, i in enumerate(range_list):
        if index == 0:
            offset_lower = 0
            offset_higher = i
        else:
            offset_lower = offset_higher
            offset_higher = offset_lower + i
        select_list.append(
            [x for x in selector.ind if (x < offset_higher) and (x >= offset_lower)]
        )
    return [mat.iloc[select] for select, mat in zip(select_list, mat_list)]



import numpy as np
import pandas as pd
from matplotlib.figure import Axes

COLOR_KEY = [
    "firebrick",
    "darkorange",
    "gold",
    "greenyellow",
    "deepskyblue",
    "navy",
    "violet",
    "magenta",
    "pink",
    "black",
    "red",
    "orange",
]


def scatter(
    data: np.ndarray | pd.DataFrame,
    ax: Axes,
    title: str,
    cell_range: dict[str, int] or None = None,
    point_size: int = 10,
):
    if isinstance(data, np.ndarray):
        if cell_range is None:
            ax.scatter(data[:, 0], data[:, 1], s=point_size, c=COLOR_KEY[0], label="data")
            ax.set_title(title)
        else:
            for i, (key, value) in enumerate(cell_range.items()):
                ax.scatter(
                    data[0 : value - 1, 0],
                    data[0 : value - 1, 1],
                    s=point_size,
                    c=COLOR_KEY[i],
                    label=key
                )
                data = data[value:]
    elif isinstance(data, pd.DataFrame):
        ax.scatter(data.iloc[:, 0], data.iloc[:, 1], s=point_size, c=COLOR_KEY[0], label="data")
    else:
        raise TypeError(f"data type {type(data)} is not supported")
    ax.set_title(title)
    ax.legend()


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()
    cell_range = {"A": 10, "B": 20, "C": 30, "D": 40, "E": 50, "F": 60, "G": 70, "H": 80, "I": 90, "J": 100}
    scatter(np.random.random((550, 2)), ax, "test", cell_range)
    fig.show()


import pandas as pd
import os

from pyRawTools import MSLoader

from .format import SCData
from .plugin import RawFileReader

from SCMeTA.config import SYSTEM


def load_txt(path):
    raw = pd.read_table(path)
    raw = raw.drop(columns=["RetentionTime"])
    raw = raw.set_index("Scan")
    return raw


def load_darwin(path):
    loader = MSLoader()
    if os.environ.get("CONTAINER"):
        temp_dir = "/ramdisk"
    else:
        temp_dir = None
    raw = loader.load(path, temp_dir=temp_dir)
    raw = raw.drop(columns=["RetentionTime"])
    return raw


def load_dotnet(path):
    file = RawFileReader(path)
    raw = file.GetFullScanMassList()
    file.Close()
    return raw


IMPORT_FUNC = {
    "Darwin": load_darwin,
    "Windows": load_dotnet,
    "Linux": load_dotnet,
}


def load_thermo_data(name, path) -> SCData:
    data = SCData(name)
    if path.lower().endswith(".raw"):
        data.raw = IMPORT_FUNC[SYSTEM](path)
    elif path.lower().endswith(".txt"):
        data.raw = load_txt(path)
    else:
        raise FileNotFoundError("File not supported")
    return data



import pandas as pd

from SCMeTA.match import Connector


class SearchDatabase:
    def __init__(self):
        pass

    @staticmethod
    def search_full(cell_mat: pd.DataFrame, **kwargs):
        mz_list = cell_mat.columns.to_list()
        with Connector() as conn:
            return conn.search_list(mz_list, **kwargs)

    def search_id(self, cell_mat: pd.DataFrame, **kwargs) -> dict[float, list[str]]:
        full_info = self.search_full(cell_mat, **kwargs)
        id = {}
        for mz, info in full_info.items():
            id[mz] = [i[1] for i in info]
        return id

    def search_name(self, cell_mat: pd.DataFrame, **kwargs) -> dict[float, list[str]]:
        full_info = self.search_full(cell_mat, **kwargs)
        name = {}
        for mz, info in full_info.items():
            name[mz] = [i[2] for i in info]
        return name



from .heatmap import heatmap
from .scatter import scatter
from .line import line
from .radar import radar
from .select import cut_range_plot
from .volcano import volcano
from .bar import bar
from .box import box


import pandas as pd
from itertools import chain


def collect_noise(mat: pd.DataFrame, cell_pos: list[list], self_sub: bool = False) -> list[pd.DataFrame]:
    noise_list = []
    for index, group in enumerate(cell_pos):
        noise_start = 1 if index == 0 else group[0] - 1
        noise_end = group[0] - 1
        noise_mean = mat.loc[noise_start:noise_end].mean()
        noise_list.append(noise_mean)
        if self_sub:
            mat.loc[noise_start:noise_end] = mat.loc[noise_start:noise_end] - noise_mean
    noise_list.append(mat.loc[cell_pos[-1][-1] + 1:].mean())
    return noise_list


def noise_subtract(
    mat: pd.DataFrame, cell_pos: list, method: str = "specific"
) -> pd.DataFrame:
    """Back subtract the noise from the raw data.
    Parameters
    ----------
    mat : pd.DataFrame
        The raw data.
    cell_pos : list
        The cell positions.
    method : str, optional
    Returns
    -------
    pd.DataFrame
        The back subtracted data.
    """
    noise_list = collect_noise(mat, cell_pos, self_sub=True)
    if method == "specific":
        for index, group in enumerate(cell_pos):
            noise_mean = pd.concat(
                [noise_list[index], noise_list[index + 1]], axis=1
            ).mean(axis=1)
            mat.loc[group] = mat.loc[group] - noise_mean
    elif method == "global":
        noise_list = list(chain.from_iterable(noise_list))
        noise_mean = mat.loc[noise_list].mean()
        for group in cell_pos:
            mat.loc[group] = mat.loc[group] - noise_mean
    mat = mat.clip(lower=0)
    return mat


def filter_assem(
    mat: pd.DataFrame,
    cell_mat: pd.DataFrame,
    cell_pos: list[list],
    snr: float = 3.0,
    method: str = "specific",
) -> pd.DataFrame:
    """Filter the assembly data.
    Parameters
    ----------
    mat : pd.DataFrame
        The raw data.
    cell_mat : pd.DataFrame
        The cell data.
    cell_pos : list[list]
        The cell positions.
    snr : float, optional
        The signal-to-noise ratio.
    method : str
        The filter method.
    Returns
    -------
    pd.DataFrame
        The filtered data.
    """
    noise_list = collect_noise(mat, cell_pos)
    mask = pd.DataFrame()
    if method == "specific":
        cell_list = cell_mat.index.tolist()
        for pos in cell_list:
            noise_filter = (
                pd.concat([noise_list[pos], noise_list[pos + 1]], axis=1).mean(axis=1)
                * snr
            )
            mask_item = cell_mat.loc[pos] > noise_filter
            mask = pd.concat([mask, mask_item], axis=1)
    elif method == "global":
        noise_list = list(chain.from_iterable(noise_list))
        noise_filter = mat.loc[noise_list].mean() * snr
        mask = pd.concat([mask, cell_mat > noise_filter], axis=1)
    mask = mask.T
    mask.index = cell_mat.index
    cell_mat = cell_mat[mask]
    # Remove the empty columns
    cell_mat = cell_mat.dropna(axis=1, how="all")
    return cell_mat



import pandas as pd

from scipy.stats import kruskal


def k_w_test(data: dict[str, pd.DataFrame]):
    np_arrays = [df.iloc[:, 0].T.to_numpy() for df in data.values()]
    h_statistic, p_value = kruskal(*np_arrays)
    return h_statistic, p_value



import numpy as np
import pandas as pd

from matplotlib.cm import ScalarMappable
import matplotlib.pyplot as plt
from matplotlib.figure import Figure, Axes
import matplotlib.ticker as ticker

from SCMeTA.method import combine_mat

CMAP_KEY = {
    "viridis": plt.cm.viridis,
    "plasma": plt.cm.plasma,
    "inferno": plt.cm.inferno,
    "magma": plt.cm.magma,
}

LOG_KEY = {
    "LOG10": np.log10,
    "LOG2": np.log2,
    "LOG": np.log,
    "NO_LOG": lambda x: x,
}


def heatmap(
    mat_list: dict[str, pd.DataFrame],
    cell_range: dict[str, int],
    ax: Axes,
    fig: Figure,
    color_map="viridis",
    func: str = "NO_LOG",
    title: str = "Heatmap",
):
    mat = combine_mat(mat_list.values()).fillna(0.00001)
    mat = mat.apply(LOG_KEY[func])
    mat = mat.T
    ax.imshow(
        mat, cmap=CMAP_KEY[color_map], aspect="auto"
    )  # aspect="auto"自动调整像素点，而非默认的正方形
    mass = mat.index.values

    cell_numb = list(cell_range.values())
    cell_name = list(cell_range.keys())

    # 横坐标以细胞名称命名
    new_cell_numb = [0] * len(cell_numb)
    for a in range(len(cell_numb)):
        if a == 0:
            new_cell_numb[a] = 0
        else:
            new_cell_numb[a] = new_cell_numb[a - 1] + cell_numb[a - 1]
    for b in range(len(cell_numb)):
        new_cell_numb[b] = new_cell_numb[b] + int(cell_numb[b] / 2)

    ax.set_xlabel("Cell")  # 设置x轴标题
    ax.set_ylabel("Mass")  # 设置y轴标题
    ax.set_title(title)  # 设置图像标题
    fig.colorbar(ScalarMappable(cmap=color_map))  # 使用color bar

    ax.xaxis.set_major_locator(ticker.FixedLocator(new_cell_numb))  # 设置x轴坐标的定位
    ax.xaxis.set_major_formatter(ticker.FixedFormatter(cell_name))  # 设置x轴坐标的名称
    ax.set_yticks(np.arange(len(mass)), labels=mass)  # 设置y轴坐标的定位和名称
    ax.yaxis.set_major_locator(ticker.MultipleLocator(30))  # 每20个y坐标显示一次



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['SCMeTA',
 'SCMeTA.accelerate',
 'SCMeTA.batch',
 'SCMeTA.config',
 'SCMeTA.core',
 'SCMeTA.file',
 'SCMeTA.file.plugin',
 'SCMeTA.match',
 'SCMeTA.method',
 'SCMeTA.method.fill',
 'SCMeTA.method.machine_learning',
 'SCMeTA.plot',
 'SCMeTA.plot.Bokeh',
 'SCMeTA.plot.Mpl',
 'SCMeTA.search']

package_data = \
{'': ['*'], 'SCMeTA.file.plugin': ['dll/*']}

install_requires = \
['combat>=0.3.3,<0.4.0',
 'matplotlib>=3.7.2,<4.0.0',
 'numpy>=1.25.2,<2.0.0',
 'pandas>=2.1.0,<3.0.0',
 'polars>=0.19.1,<0.20.0',
 'pyrawtools>=0.2.3,<0.3.0',
 'pythonnet>=3.0.3,<4.0.0',
 'scikit-learn>=1.3.0,<2.0.0',
 'scipy>=1.11.2,<2.0.0',
 'tqdm>=4.66.1,<5.0.0',
 'umap-learn>=0.5.4,<0.6.0']

setup_kwargs = {
    'name': 'scmeta',
    'version': '0.2.2',
    'description': 'A python package for single-cell metabolism analysis.',
    'long_description': '# SCMeTA\n\nSCMeTA is a python library of single-cell meta-analysis tools. It provides a set of functions for single-cell meta-analysis, including data integration, batch effect correction, cell type annotation, cell clustering, cell trajectory inference, and cell type marker identification. It also provides a set of functions for single-cell data visualization, including dimension reduction, cell clustering, cell trajectory inference, and cell type marker identification. \n\n## Installation\n\nSCMeTA is available on PyPI and can be installed with pip:\n\n```bash\npip install scmeta\n```\n\n## Usage\n\n### Data integration\n\n```python\n\nfrom SCMeTA import Process\n\nsc = Process()\n\n# Load data\n\nsc.load("data/example.RAW")\n\n# Data process\n\nsc.pre_process()\nsc.process()\nsc.post_process()\n\n```\n\n## Documentation\n\nThe official documentation is hosted on Read the Docs: https://sc-meta.com/\n\n## License\n\nSCMeTA is licensed under the GPLv3 license. See the LICENSE file for more details.\n\n\n',
    'author': 'EstrellaXD',
    'author_email': 'estrellaxd05@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.sc-meta.org/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.13',
}


setup(**setup_kwargs)



DEFAULT_CONFIG = """
[DEFAULT]
[PARAMETERS]
Resolution = 0.01
Count = 10
MaxRatio = 0.1
Adjacent = 3
Snr = 3.0
Threshold = 0.2
Lock = false


[METABOLITE]
Online = False
Include = config/include.csv
Exclude = config/exclude.csv
"""


from math import pi

import pandas as pd
import numpy as np

from bokeh.models import BasicTicker, ColorBar, LinearColorMapper, PrintfTickFormatter
from bokeh.plotting import figure, show


def heatmap(mat_list: list[pd.DataFrame], title: str = "Heatmap"):
    """
    Plot a heatmap of the data

    Args:
        mat_list: list of dataframes
        title: title of the plot

    Returns:
        bokeh figure
    """
    mat = pd.concat(mat_list, axis=0).fillna(1)
    mat = mat.apply(np.log10)
    mat.reset_index(drop=True, inplace=True)
    mat["Scan"] = mat["Scan"].astype(str)
    mat = mat.set_index("Scan")
    mat.columns.name = "Mass"

    scan = mat.index
    mass = mat.columns

    df = pd.DataFrame(mat.stack(), columns=["Intensity"]).reset_index()

    mapper = LinearColorMapper(
        palette="Viridis256", low=df.Intensity.min(), high=df.Intensity.max()
    )

    TOOLS = "hover,save,pan,box_zoom,reset,wheel_zoom"

    p = figure(
        title=title,
        x_range=list(scan),
        y_range=list(mass),
        x_axis_location="above",
        width=900,
        height=900,
        tools=TOOLS,
        toolbar_location="below",
    )

    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "5pt"
    p.axis.major_label_standoff = 0
    p.xaxis.major_label_orientation = pi / 3

    p.rect(
        x="Scan",
        y="Mass",
        width=1,
        height=1,
        source=df,
        fill_color={"field": "Intensity", "transform": mapper},
        line_color=None,
    )

    color_bar = ColorBar(
        color_mapper=mapper,
        major_label_text_font_size="5pt",
        ticker=BasicTicker(desired_num_ticks=len(df.Intensity.unique())),
        formatter=PrintfTickFormatter(format="%d"),
        label_standoff=6,
        border_line_color=None,
        location=(0, 0),
    )

    p.add_layout(color_bar, "right")

    show(p)
